from solo_turnier import worker, html_parser
import os
import json
import pytest
import pytest_mock

def __importJSONData(name):
    path = os.path.join(os.path.dirname(__file__), 'worker', name)
    with open(path, 'r') as fp:
        return json.load(fp)

@pytest.fixture
def fixture_csvExtractor():
    data = __importJSONData('csvImport.json')
    expected = __importJSONData('csvImportResult.json')
    return (data, expected)

def test_csvExtractor(fixture_csvExtractor):
    extractor = worker.CSVExtractor()
    mapped = extractor.mapCSVImport(fixture_csvExtractor[0])

    assert len(mapped) == len(fixture_csvExtractor[1])
    for i,elem in enumerate(fixture_csvExtractor[1]):
        assert mapped[i].__dict__ == elem

def test_extractPersonFromRow():
    row = worker.ResultRow('Max', 'Mustermann', 'TSC Entenhausen', '2', 'Kin', 'Adv.', 'Rumba', '2', '2', 'Kin.', 'Beg./Adv.')
    person = worker.ResultPerson.extractFromResultRow(row)
    expected = {
        'firstName': 'Max',
        'lastName': 'Mustermann',
        'name': 'Max Mustermann',
        'club': 'TSC Entenhausen',
        'id': None,
        'group': None
    }
    assert person.__dict__ == expected

def test_extractCompetitionFromRow():
    row = worker.ResultRow('Max', 'Mustermann', 'TSC Entenhausen', '2', 'Kin', 'Adv.', 'Rumba', '2', '2', 'Kin.', 'Beg./Adv.')
    competition = worker.CompetitionResult.extractFromResultRow(row)
    expected = {
        'dance': 'Rumba',
        'class_': 'Adv.', 
        'group': 'Kin',
        'place': '2',
        'placeTo': '2',
        'id': 2,
        'finalist': None,
        'competitionGroup': 'Kin.',
        'competitionClass': 'Beg./Adv.'
    }
    assert competition.__dict__ == expected

def test_combineRowsByPerson():
    rows = [
        worker.ResultRow('Max', 'Mustermann', 'TSC Entenhausen', '2', 'Kin', 'Adv.', 'Cha Cha', '-', '-', 'Kin.', 'Adv.'),
        worker.ResultRow('Max', 'Mustermann', 'TSC Entenhausen', '2', 'Kin', 'Adv.', 'Rumba', '2', '2', 'Kin.', 'Adv.'),
        worker.ResultRow('Max', 'Mustermann', 'TSC Entenhausen', '2', 'Kin', 'Beg.', 'Jive', '1', '1', 'Kin.', 'Beg.'),
        worker.ResultRow('Maxime', 'Musterfrau', '1. SC Entenhausen', '1', 'Kin', 'Adv.', 'Rumba', '1', '1', 'Kin.', 'Adv.')
    ]
    dataWorker = worker.DataWorker()
    result = dataWorker.combineRowsByPerson(rows)
    expected = {
        worker.ResultPerson('Max', 'Mustermann', 'TSC Entenhausen'): [
            worker.CompetitionResult('Rumba', 'Kin', 'Adv.', '2', '2', '2', 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin', 'Beg.', '1', '1', '2', 'Kin.', 'Beg.')
        ],
        worker.ResultPerson('Maxime', 'Musterfrau', '1. SC Entenhausen'): [
            worker.CompetitionResult('Rumba', 'Kin', 'Adv.', '1', '1', '1', 'Kin.', 'Adv.')
        ]
    }
    assert result == expected

def test_checkUniqueIds_True():
    person1 = worker.ResultPerson('Max', 'Mustermann', 'TSC Entenhausen')
    person2 = worker.ResultPerson('Maxime', 'Musterfrau', '1. SC Entenhausen')
    data = {
        person1: [
            worker.CompetitionResult('Rumba', 'Kin', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin', 'Beg.', '1', '1', 2, 'Kin.', 'Beg.')
        ],
        person2: [
            worker.CompetitionResult('Rumba', 'Kin', 'Adv.', '1', '1', 1, 'Kin.', 'Adv.')
        ]
    }
    dataWorker = worker.DataWorker()
    assert dataWorker.checkUniqueIds(data) == True
    assert person1.id == 2
    assert person2.id == 1

def test_checkUniqueIds_False():
    person1 = worker.ResultPerson('Max', 'Mustermann', 'TSC Entenhausen')
    person2 = worker.ResultPerson('Maxime', 'Musterfrau', '1. SC Entenhausen')
    data = {
        person1: [
            worker.CompetitionResult('Rumba', 'Kin', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin', 'Beg.', '1', '1', 3, 'Kin.', 'Beg.')
        ],
        person2: [
            worker.CompetitionResult('Rumba', 'Kin', 'Adv.', '1', '1', 1, 'Kin.', 'Adv.')
        ]
    }
    dataWorker = worker.DataWorker()
    assert dataWorker.checkUniqueIds(data) == False
    assert person1.id == None
    assert person2.id == 1

@pytest.fixture(params=range(5))
def fixture_consolidateGroups(request):
    person1 = worker.ResultPerson('Max 1', 'Mustermann', 'TSC Entenhausen')
    person2 = worker.ResultPerson('Max 2', 'Mustermann', 'TSC Entenhausen')
    person3 = worker.ResultPerson('Max 3', 'Mustermann', 'TSC Entenhausen')
    person4 = worker.ResultPerson('Max 4', 'Mustermann', 'TSC Entenhausen')
    # persons = (person1, person2, person3, person4)

    dict1 = {
        person1: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin.', 'Beg.', '1', '1', 3, 'Kin.', 'Beg.')
        ]
    }
    dict2 = {
        person2: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin./Jun.', 'Beg.', '1', '1', 3, 'Kin./Jun.', 'Beg.')
        ]
    }
    dict3 = {
        person3: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.')
        ]
    }
    dict4 = {
        person4: [
            worker.CompetitionResult('Rumba', 'Kin./Jun.', 'Adv.', '2', '2', 2, 'Kin./Jun.', 'Adv.')
        ]
    }
    dict5 = {
        person4: [
            worker.CompetitionResult('Rumba', 'Kin./Jun.', 'Adv.', '2', '2', 2, 'Kin./Jun.', 'Adv.'),
            worker.CompetitionResult('Cha Cha', 'Jun./Jug.', 'Beg.', '3', '4', 2, 'Jun./Jug.', 'Beg.')
        ]
    }

    cases = (
        (dict1|dict3, (True, False), {}),
        (dict1|dict2|dict3, (True, True), {}),
        (dict4, (False, False), {person4: 'Kin./Jun.'}),
        (dict1|dict2|dict3|dict4, (False, True), {person4: 'Kin./Jun.'}),
        (dict5, (True, True), {person4: 'Jun.'}),
    )
    return cases[request.param]

@pytest.fixture(params=range(2))
def fixture_consolidateGroups_fail(request, fixture_consolidateGroups):
    person = worker.ResultPerson('Max 5', 'Mustermann', 'TSC Entenhausen')

    dict1 = {
        person: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Jun.', 'Beg.', '1', '1', 3, 'Jun.', 'Adv.')
        ]
    }

    dict2 = {
        person: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Hgr', 'Beg.', '1', '1', 3, 'Hgr', 'Adv.')
        ]
    }

    cases = (
        dict1 | fixture_consolidateGroups[0],
        dict2 | fixture_consolidateGroups[0]
    )
    return cases[request.param]

def test_consolidateGroups(fixture_consolidateGroups):
    data = fixture_consolidateGroups[0]
    dataWorker = worker.DataWorker()

    assert dataWorker.consolidateGroups(data) == fixture_consolidateGroups[1]

    for person in data:
        assert person.group == fixture_consolidateGroups[2].get(person, 'Kin.')

def test_consolidateGroups_failing(fixture_consolidateGroups_fail):
    data = fixture_consolidateGroups_fail
    dataWorker = worker.DataWorker()

    with pytest.raises(Exception):
        dataWorker.consolidateGroups(data)

def test_createHtmlLUT(mocker):
    mock = mocker.patch('solo_turnier.html_parser.HtmlParser.guessDataFromHtmlTitle')
    mock.side_effect= [
        {'group': 'group1', 'class_': 'class1', 'dance': 'dance1'},
        {'group': 'group2', 'class_': 'class2', 'dance': 'dance2'},
        {'group': 'group3', 'class_': 'class3', 'dance': 'dance3'},
    ]
    
    importMock1 = mocker.patch('solo_turnier.html_parser.HtmlImport')
    importMock2 = mocker.patch('solo_turnier.html_parser.HtmlImport')
    importMock3 = mocker.patch('solo_turnier.html_parser.HtmlImport')
    importMock1.title = 'Fake title 1'
    importMock2.title = 'Fake title 2'
    importMock3.title = 'Fake title 3'

    dataWorker = worker.DataWorker()
    structure = dataWorker._createHtmlLUT([importMock1, importMock2, importMock3])

    expected = {
        ('group1', 'class1', 'dance1'): importMock1,
        ('group2', 'class2', 'dance2'): importMock2,
        ('group3', 'class3', 'dance3'): importMock3,
    }
    assert expected == structure

def test_mergeHtmlData(mocker):
    person1 = worker.ResultPerson('Max 1', 'Mustermann', 'TSC Entenhausen')
    person2 = worker.ResultPerson('Max 2', 'Mustermann', 'TSC Entenhausen')
    person3 = worker.ResultPerson('Max 3', 'Mustermann', 'TSC Entenhausen')
    person4 = worker.ResultPerson('Max 4', 'Mustermann', 'TSC Entenhausen')

    data = {
        person1: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Beg.', '1', '1', 1, 'Kin./Jun.', 'Beg.'),
            worker.CompetitionResult('Cha Cha', 'Kin.', 'Adv.', '1', '1', 1, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin.', 'Beg.', '1', '2', 1, 'Kin.', 'Beg.'),
            worker.CompetitionResult('Langs. Walzer', 'Kin.', 'Beg.', '1', '1', 1, 'Kin.', 'Newc./Beg.'),
        ],
        person2: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Beg.', '2', '2', 2, 'Kin./Jun.', 'Beg.'),
            worker.CompetitionResult('Cha Cha', 'Kin.', 'Adv.', '2', '2', 2, 'Kin.', 'Adv.'),
            worker.CompetitionResult('Jive', 'Kin.', 'Beg.', '1', '2', 2, 'Kin.', 'Beg.'),
            worker.CompetitionResult('Langs. Walzer', 'Kin.', 'Newc.', '1', '1', 2, 'Kin.', 'Newc./Beg.'),
        ],
        person3: [
            worker.CompetitionResult('Rumba', 'Jun.', 'Beg.', '1', '1', 3, 'Kin./Jun.', 'Beg.'),
            # worker.CompetitionResult('Cha Cha', 'Jun.', 'Adv.', '1', '1', 3, 'Kin.', 'Adv.'),
            # worker.CompetitionResult('Jive', 'Jun.', 'Beg.', '2', '2', 3, 'Kin.', 'Beg.'),
            # worker.CompetitionResult('Langs. Walzer', 'Jun.', 'Newc./Beg.', '1', '1', 3, 'Kin.', 'Beg.'),
        ],
        person4: [
            worker.CompetitionResult('Rumba', 'Kin.', 'Beg.', '3', '3', 4, 'Kin./Jun.', 'Beg.'),
            # worker.CompetitionResult('Cha Cha', 'Kin.', 'Adv.', '1', '1', 4, 'Kin.', 'Adv.'),
            # worker.CompetitionResult('Jive', 'Kin.', 'Beg.', '2', '2', 4, 'Kin.', 'Beg.'),
            # worker.CompetitionResult('Langs. Walzer', 'Kin.', 'Newc./Beg.', '1', '1', 4, 'Kin.', 'Beg.'),
        ],
    }

    htmlParticipant1Dance1 = html_parser.HtmlParticipant('Max 1 Mustermann', '1.', True)
    htmlParticipant1Dance2 = html_parser.HtmlParticipant('Max 1 Mustermann', '1.', True)
    htmlParticipant1Dance3 = html_parser.HtmlParticipant('Max 1 Mustermann', '1.-2.', True)
    htmlParticipant1Dance4 = html_parser.HtmlParticipant('Max 1 Mustermann', '1.', True)
    
    htmlParticipant2Dance1 = html_parser.HtmlParticipant('Max 2 Mustermann', '2.', True)
    htmlParticipant2Dance2 = html_parser.HtmlParticipant('Max 2 Mustermann', '2.', True)
    htmlParticipant2Dance3 = html_parser.HtmlParticipant('Max 2 Mustermann', '1.-2.', True)
    htmlParticipant2Dance4 = html_parser.HtmlParticipant('Max 2 Mustermann', '1.', True)
    
    htmlParticipant3Dance1 = html_parser.HtmlParticipant('Max 3 Mustermann', '1.', True)

    htmlParticipant4Dance1 = html_parser.HtmlParticipant('Max 4 Mustermann', '3.', False)

    htmlParticipantsDance1 = {
        '1': htmlParticipant1Dance1,
        '2': htmlParticipant2Dance1,
        '3': htmlParticipant3Dance1,
        '4': htmlParticipant4Dance1
    }
    htmlParticipantsDance2 = {
        '1': htmlParticipant1Dance2,
        '2': htmlParticipant2Dance2,
    }
    htmlParticipantsDance3 = {
        '1': htmlParticipant1Dance3,
        '2': htmlParticipant2Dance3,
    }
    htmlParticipantsDance4 = {
        '1': htmlParticipant1Dance4,
        '2': htmlParticipant2Dance4,
    }

    htmlCompetition1 = html_parser.HtmlImport('ETW, Solos Kin./Jun. Beginner Rumba', htmlParticipantsDance1)
    htmlCompetition2 = html_parser.HtmlImport('ETW, Solos Kin. Advanced Cha Cha', htmlParticipantsDance2)
    htmlCompetition3 = html_parser.HtmlImport('ETW, Solos Kinder Beginner Jive', htmlParticipantsDance3)
    htmlCompetition4 = html_parser.HtmlImport('ETW, Solos Kin. Newc./Beg. Langs. Walzer', htmlParticipantsDance4)
    
    dataWorker = worker.DataWorker()

    dataWorker.mergeHtmlData(data, [htmlCompetition1, htmlCompetition2, htmlCompetition3, htmlCompetition4])

    person1Finalist = [c.finalist for c in data[person1]]
    person2Finalist = [c.finalist for c in data[person2]]
    person3Finalist = [c.finalist for c in data[person3]]
    person4Finalist = [c.finalist for c in data[person4]]
    finalists = {
        person1: person1Finalist,
        person2: person2Finalist,
        person3: person3Finalist,
        person4: person4Finalist,
    }
    expectedFinalists = {
        person1: [True, True, True, True],
        person2: [True, True, True, True],
        person3: [True],
        person4: [False],    
    }

    assert finalists == expectedFinalists

@pytest.fixture(params=range(4))
def fixture_getAllDancesInCompetition(request, mocker):
    def mockCompetition(comp):
        def mockUser():
            return mocker.patch('solo_turnier.worker.ResultPerson')
        def mockDances(dances):
            def mockDance(name):
                mock = mocker.patch('solo_turnier.worker.CompetitionResult')
                mock.dance = name
                return mock
            return [mockDance(d) for d in dances]
        return {mockUser(): mockDances(dances) for dances in comp}
    
    cases = (
        ([['Samba']], ['Samba']),
        ([['Samba', 'Rumba'], ['Cha Cha']], ['Samba', 'Cha Cha', 'Rumba']),
        ([['Samba', 'Rumba'], ['Cha Cha', 'Tango', 'Langs. Walzer']], ['Samba', 'Cha Cha', 'Rumba', 'Langs. Walzer', 'Tango']),
        ([['Cha Cha', 'Rumba', 'Jive'], ['Quickstep', 'Tango', 'Wiener Walzer', 'Langs. Walzer'], ['Slowfox', 'Langs. Walzer', 'Paso Doble', 'Samba']], ['Samba', 'Cha Cha', 'Rumba', 'Paso Doble', 'Jive', 'Langs. Walzer', 'Tango', 'Wiener Walzer', 'Slowfox', 'Quickstep'])
    )
    case = cases[request.param]
    return (mockCompetition(case[0]), case[1])

def test_getAllDancesInCompetitions(fixture_getAllDancesInCompetition):
    print(fixture_getAllDancesInCompetition)
    data = fixture_getAllDancesInCompetition[0]
    dataWorker = worker.DataWorker()
    ret = dataWorker.getAllDancesInCompetitions(data)
    assert ret == fixture_getAllDancesInCompetition[1]

def test_collectPersonsInGroups(mocker):
    def mockPerson(group):
        mock = mocker.patch('solo_turnier.worker.ResultPerson')
        mock.group = group
        return mock
    persons = (
        mockPerson('Kin.'), mockPerson('Kin.'), mockPerson('Jun.'),
        mockPerson('Kin.'), mockPerson(None), mockPerson('Jug.'),
        mockPerson(None), mockPerson('Kin./Jun.'), mockPerson('Jun.')
    )
    data = {p: [] for p in persons}
    dataWorker = worker.DataWorker()
    groups = dataWorker.collectPersonsInGroups(data)

    assert groups['Kin.'] == [persons[0], persons[1], persons[3]]
    assert groups['Jun.'] == [persons[2], persons[8]]
    assert groups['Jug.'] == [persons[5]]
    assert groups['Sonst'] == [persons[4], persons[6], persons[7]]

def test_sortPersons_withId(mocker):
    def mockPerson(id):
        mock = mocker.patch('solo_turnier.worker.ResultPerson')
        mock.id = id
        return mock
    persons = [mockPerson(2), mockPerson(1), mockPerson(5), mockPerson(3)]
    dataWorker = worker.DataWorker()
    sorted, showIds = dataWorker.sortPersonsInGroup(persons)
    assert sorted == [persons[1], persons[0], persons[3], persons[2]]
    assert showIds == True

def test_sortPersons_withoutId(mocker):
    def mockPerson(name):
        mock = mocker.patch('solo_turnier.worker.ResultPerson')
        mock.id = 3
        mock.name = name
        mock.club = 'TSC Entenhausen'
        return mock
    persons = [mockPerson('Max'), mockPerson('Isabel'), mockPerson('Reimund'), mockPerson('Anna')]
    persons[2].id = None
    dataWorker = worker.DataWorker()
    sorted, showIds = dataWorker.sortPersonsInGroup(persons)
    assert sorted == [persons[3], persons[1], persons[0], persons[2]]
    assert showIds == False

def test_mapPersonResultsToDanceList(mocker):
    def mockResult(dance):
        mock = mocker.patch('solo_turnier.worker.CompetitionResult')
        mock.dance = dance
        return mock
    dances = ['Cha Cha', 'Rumba', 'Langs. Walzer', 'Quickstep']
    results = [mockResult('Rumba'), mockResult('Quickstep'), mockResult('Cha Cha')]
    dataWorker = worker.DataWorker()
    mappedResults = dataWorker.mapPersonResultsToDanceList(results, dances)
    assert mappedResults == [results[2], results[0], None, results[1]]
